;; Matrix multiplication using N^2 processors, for N = 2^k
;; Geometry should be N by N
;; File:  mmpy2.lisp

(in-package '*lisp)

(*proclaim '(type single-float-pvar A B C U V R))

(*defvar A (!! 0.0))	;; Three matrices, N by N
(*defvar B (!! 0.0))
(*defvar C (!! 0.0))
(*defvar U (!! 0.0))	;; Working pvars
(*defvar V (!! 0.0))
(*defvar R (!! 0.0))


(*proclaim '(type fixnum-pvar i j))

(*defvar i (!! 0))	; Will be grid coord 0
(*defvar j (!! 1))	; Will be grid coord 1
(*defvar k (!! 0))	; Used in address calculations

(*proclaim '(type fixnum nx))

(defvar nx 4)		;; Matrix size for current tests

;; mult4 takes arguments X, Y, returns result pvar such that
;; Result[*,*] = X[*,*]*Y[*,*] (matrix product).  It expects
;; to be called with all processors active.

(*defun mult4 (X Y)
 (declare (type single-float-pvar X Y))
  (*set U X)  (*set V Y)
 ;; In U, rotate row i left i positions
  (*set k (+!! j (-!! (!! nx) i)))
  (*if (>=!! k (!! nx)) (*set k (-!! k (!! nx))))
  (*when (plusp!! i) (*pset :no-collisions X U (cube-from-grid-address!! i k)))
 ;; In V, rotate column j up j positions
  (*set k (+!! i (-!! (!! nx) j)))
  (*if (>=!! k (!! nx)) (*set k (-!! k (!! nx))))
  (*when (plusp!! j) (*pset :no-collisions Y V (cube-from-grid-address!! k j)))
 ;; Form product
  (*set R (*!! U V))
  (dotimes (kx (1- nx))
    (*set U (news!! U 0 -1))  (*set V (news!! V -1 0))
    (*set R (+!! R (*!! U V)))
  ) ; end dotimes
  R	;; Result
)

;; Test driver for mult4, times n multiplications

(defun test4 (n)
  (cm:time (test4x n))
)

(defun test4x (n)
  (dotimes (tk n)
    (*set C (mult4 A B))
  ) ; end dotimes
  (setq flops (* n flm))
)

;; Initialize for N by N matrices (N a power of 2)

(defvar flm    128)	;; Flops per multiplication = 2*N^3
(defvar flops 1280)	;; Flops in most recent experiment

(defun setup (n)
  (*cold-boot :initial-dimensions (list n n))
  (setq nx n)
  (setq flm (* 2 nx nx nx))
  (*set A (frndm!! 5.0))
  (*set B (frndm!! 5.0))
  (*set i (self-address-grid!! (!! 0)))
  (*set j (self-address-grid!! (!! 1)))
  (format T "~%Initialized for ~D by ~D matrices.~%" nx nx)
)

;; Returns random single-float-pvar with values uniformly distributed
;; between -rng and rng.

(*defun frndm!! (rng)
  (*let ((rn (!! 0.0)))
	(declare (type single-float-pvar rn))
    (*set rn (random!! (!! (+ rng rng))))
    (-!! rn (!! rng))
  )
)

;; Returns identity matrix

(*defun ident ()
  (if!! (=!! i j) (!! 1.0) (!! 0.0))
)

;; Pretty-print matrix from X

(*defun ppm (X)
  (ppp X :format "~8,3F" :mode :grid :ordering '(1 0))
)
